import java.io.IOException;
import java.io.InputStream;

import ch.ethz.ssh2.Connection;
import ch.ethz.ssh2.ServerHostKeyVerifier;
import ch.ethz.ssh2.Session;

/**
 * SSH2Demo2. A simple demo that shows how one can read from stdout and stderr
 * in parallel without using two threads (and without risking a deadlock).
 * 
 * @author Christian Plattner, plattner@inf.ethz.ch
 * @version $Id: SSH2Demo2.java,v 1.2 2005/06/27 07:33:40 cplattne Exp $
 * 
 */
public class SSH2Demo2 implements ServerHostKeyVerifier
{

	public void work(String host, int port, String user, String password)
	{
		Connection sshc = new Connection(host, port);

		try
		{
			/* Connect and register callback */

			sshc.connect(this);

			/* Authenticate */

			boolean res = sshc.authenticateWithPassword(user, password);

			if (res == false)
			{
				throw new IOException("Could not authenticate!");
			}

			/* Create a session */

			Session sess = sshc.openSession();

			/*
			 * Do NOT Request a PTY and execute something.
			 */

			sess.execCommand("echo \"Text on STDOUT\"; echo \"Text on STDERR\" >&2");

			/*
			 * The following is a demo on how one can read from stdout and
			 * stderr without having to use two parallel worker threads.
			 */

			InputStream stdout = sess.getStdout();
			InputStream stderr = sess.getStderr();

			while (true)
			{
				int stdoutAvail = stdout.available();
				int stderrAvail = stderr.available();

				if ((stdoutAvail == 0) && (stderrAvail == 0))
				{
					/* wait no longer than 2 seconds (= 2000 milliseconds) */

					int status = sess.waitUntilDataAvailable(2000);

					if (status < 0)
					{
						/* A timeout occured. */
						throw new IOException("Timeout while waiting for data from peer.");
					}

					if (status == 0)
					{
						/*
						 * There is no more data, EOF reached on both stdout and
						 * stderr.
						 */
						break;
					}

					/* Since status > 0, there is more data vailable */

					stdoutAvail = stdout.available();
					stderrAvail = stderr.available();
				}

				while (stdoutAvail > 0)
				{
					int c = stdout.read();
					if (c == -1) // you never know
						break;
					System.out.print((char) c);
					stdoutAvail--;
				}

				while (stderrAvail > 0)
				{
					int c = stderr.read();
					if (c == -1) // you never know
						break;
					System.err.print((char) c);
					stderrAvail--;
				}
			}

			/* Show exit status, if available */

			System.out.println("ExitCode: " + sess.getExitStatus());

		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		finally
		{
			/* Close the connection (=> stop the receiver thread) */

			sshc.close();
		}
	}

	public boolean verifyServerHostKey(String hostname, int port, String serverHostKeyAlgorithm, byte[] serverHostKey)
	{
		System.out.println("Checking key from " + hostname + ":" + port);

		/* Should keep a database of already collected keys and always compare */

		return true;
	}

	public static void main(String[] args)
	{
		SSH2Demo2 demo2 = new SSH2Demo2();
		demo2.work("HOST", 22, "USER", "PASSWORD");
	}
}
