
package ch.ethz.ssh2.transport;

import java.io.IOException;
import java.net.Socket;
import java.util.Vector;

import ch.ethz.ssh2.ConnectionInfo;
import ch.ethz.ssh2.DHGexParameters;
import ch.ethz.ssh2.ServerHostKeyVerifier;
import ch.ethz.ssh2.crypto.CryptoWishList;
import ch.ethz.ssh2.crypto.cipher.BlockCipher;
import ch.ethz.ssh2.crypto.digest.MAC;
import ch.ethz.ssh2.packets.Packets;

/*
 * Yes, the "standard" is a big mess. On one side, the say that arbitary channel
 * packets are allowed during kex exchange, on the other side we need to blindly
 * ignore the next _packet_ if the KEX guess was wrong. Where do we know from that
 * the next packet is not a channel data packet? Yes, we could check if it is in
 * the KEX range. But the standard says nothing about this. The OpenSSH guys
 * block local "normal" traffic during KEX. That's fine - however, they assume
 * that the other side is doing the same. During re-key, if they receive traffic
 * other than KEX, they become horribly irritated and kill the connection. Since
 * we are very likely going to communicate with OpenSSH servers, we have to play
 * the same stupid game - even though we could do better.
 * 
 * btw: having stdout and stderr on the same channel, with a shared window, is
 * also a VERY good idea... =(
 */

/**
 * TransportManager.
 * 
 * @author Christian Plattner, plattner@inf.ethz.ch
 * @version $Id: TransportManager.java,v 1.3 2005/06/06 12:44:23 cplattne Exp $
 */
public class TransportManager
{
	class HandlerEntry
	{
		MessageHandler mh;
		int low;
		int high;
	}

	String hostname;
	int port;
	Socket sock;

	Object connectionSemaphore = new Object();

	boolean flagKexOngoing = false;
	boolean connectionClosed = false;
	String reasonClosed;

	Vector kexBlockedSendQueue = new Vector(10);

	TransportConnection tc;
	KexManager km;

	Vector messageHandlers = new Vector();

	Thread receiveThread;

	public TransportManager(String host, int port) throws IOException
	{
		this.hostname = host;
		this.port = port;

		try
		{
			sock = new Socket(host, port);
		}
		catch (IOException e)
		{
			throw new IOException("Cannot establish TCP connection to " + host + ":" + port);
		}

		tc = new TransportConnection(sock.getInputStream(), sock.getOutputStream());
	}

	public ConnectionInfo getConnectionInfo(int kexNumber) throws IOException
	{
		return km.getOrWaitForConnectionInfo(kexNumber);
	}

	public String getReasonClosed()
	{
		synchronized (connectionSemaphore)
		{
			return reasonClosed;
		}
	}

	public byte[] getSessionIdentifier()
	{
		return km.sessionId;
	}

	public void close(String reason)
	{
		synchronized (connectionSemaphore)
		{
			if (connectionClosed)
				return;

			try
			{
				sock.close();
			}
			catch (IOException ignore)
			{
			}

			connectionClosed = true;
			reasonClosed = reason;

			connectionSemaphore.notifyAll();
		}
	}

	public void initialize(CryptoWishList cwl, ServerHostKeyVerifier verifier, DHGexParameters dhgex) throws IOException
	{
		ClientServerHello csh = new ClientServerHello(sock.getInputStream(), sock.getOutputStream());

		km = new KexManager(this, csh, cwl, hostname, port, verifier);
		km.initiateKEX(cwl, dhgex);

		receiveThread = new Thread(new Runnable()
		{
			public void run()
			{
				try
				{
					receiveLoop();
				}
				catch (IOException e)
				{
					close(e.getMessage());
				}

				if (km != null)
				{
					try
					{
						km.handleMessage(null, 0);
					}
					catch (IOException e)
					{
					}
				}

				for (int i = 0; i < messageHandlers.size(); i++)
				{
					HandlerEntry he = (HandlerEntry) messageHandlers.elementAt(i);
					try
					{
						he.mh.handleMessage(null, 0);
					}
					catch (Exception ignore)
					{
					}
				}

			}
		});

		receiveThread.start();
	}

	public void registerMessageHandler(MessageHandler mh, int low, int high)
	{
		HandlerEntry he = new HandlerEntry();
		he.mh = mh;
		he.low = low;
		he.high = high;

		synchronized (messageHandlers)
		{
			messageHandlers.addElement(he);
			/* TODO: check for multiple entries */
		}
	}

	public void removeMessageHandler(MessageHandler mh, int low, int high)
	{
		synchronized (messageHandlers)
		{
			for (int i = 0; i < messageHandlers.size(); i++)
			{
				HandlerEntry he = (HandlerEntry) messageHandlers.elementAt(i);
				if ((he.mh == mh) && (he.low == low) && (he.high == high))
				{
					messageHandlers.removeElementAt(i);
					/* TODO: check for multiple entries */
					break;
				}
			}
		}
	}

	public void sendKexMessage(byte[] msg) throws IOException
	{
		synchronized (connectionSemaphore)
		{
			if (connectionClosed)
			{
				String err = "Sorry, this connection is closed.";
				if (reasonClosed != null)
					err += (" (" + reasonClosed + ")");
				throw new IOException(err);
			}

			flagKexOngoing = true;

			try
			{
				tc.sendMessage(msg);
			}
			catch (IOException e)
			{
				close(e.getMessage());
				throw e;
			}
		}
	}

	public void kexFinished() throws IOException
	{
		synchronized (connectionSemaphore)
		{
			flagKexOngoing = false;

			for (int i = 0; i < kexBlockedSendQueue.size(); i++)
			{
				tc.sendMessage((byte[]) kexBlockedSendQueue.elementAt(i));
			}

			connectionSemaphore.notifyAll();
		}
	}

	public void forceKeyExchange(CryptoWishList cwl, DHGexParameters dhgex) throws IOException
	{
		km.initiateKEX(cwl, dhgex);
	}

	public void changeRecvCipher(BlockCipher bc, MAC mac)
	{
		tc.changeRecvCipher(bc, mac);
	}

	public void changeSendCipher(BlockCipher bc, MAC mac)
	{
		tc.changeSendCipher(bc, mac);
	}

	public void sendMessage(byte[] msg) throws IOException
	{
		synchronized (connectionSemaphore)
		{
			while (true)
			{
				if (connectionClosed)
				{
					String err = "Sorry, this connection is closed.";
					if (reasonClosed != null)
						err += (" (" + reasonClosed + ")");
					throw new IOException(err);
				}

				if (flagKexOngoing == false)
					break;

				if (Thread.currentThread() == receiveThread)
				{
					// Never block the receive thread
					// Use a queue for non-KEX reply packets
					// which were generated in the receive thread
					kexBlockedSendQueue.addElement(msg);
					return;
				}

				try
				{
					connectionSemaphore.wait();
				}
				catch (InterruptedException e)
				{
				}
			}

			try
			{
				// Packets.debug("SEND: ", msg);
				tc.sendMessage(msg);
			}
			catch (IOException e)
			{
				close(e.getMessage());
				throw e;
			}
		}
	}

	public void receiveLoop() throws IOException
	{
		byte[] msg = new byte[35000];

		while (true)
		{
			int msglen = tc.receiveMessage(msg, 0, msg.length);

			// System.out.println("REC: MSG " + buff[0]);

			int type = msg[0] & 0xff;

			if (type == Packets.SSH_MSG_IGNORE)
				continue;

			if (type == Packets.SSH_MSG_DEBUG)
				continue;

			if (type == Packets.SSH_MSG_UNIMPLEMENTED)
			{
				throw new IOException("Peer sent UNIMPLEMENTED message, that should not happen.");
			}

			if (type == Packets.SSH_MSG_DISCONNECT)
			{
				// SSHPackets.debug("DISC: ", msg);
				throw new IOException("Peer sent DISCONNECT message.");
			}

			/*
			 * Is it a KEX Packet?
			 */

			if ((type == Packets.SSH_MSG_KEXINIT) || (type == Packets.SSH_MSG_NEWKEYS)
					|| ((type >= 30) && (type <= 49)))
			{
				km.handleMessage(msg, msglen);
				continue;
			}

			MessageHandler mh = null;

			for (int i = 0; i < messageHandlers.size(); i++)
			{
				HandlerEntry he = (HandlerEntry) messageHandlers.elementAt(i);
				if ((he.low <= type) && (type <= he.high))
				{
					mh = he.mh;
					break;
				}

			}

			if (mh == null)
				throw new IOException("Unexpected SSH message (type " + type + ")");

			mh.handleMessage(msg, msglen);
		}
	}
}
